@@ -15,6 +15,7 @@ use POSIX;
 
 my $hostname = `hostname`;
 $hostname =~ s/\n//g;
+
 # configfile is the only thing you should set. Other stuff is overriden by
 # contents of the config file
 my $configfile = '/etc/viperdb.conf';
@@ -33,7 +34,7 @@ my %ignore;
 my @rignore;
 my $md5obj = new MD5;
 
-#Grab da OS name
+# Grab da OS name
 chomp($os_type=lc(qx[uname]));
 #open LOG, "|cat";
 
@@ -41,6 +42,11 @@ chomp($os_type=lc(qx[uname]));
 
 &Lock;
 
+# logfile is only used if defined in config file.
+if (defined $logfile) {
+	open LOG, ">>$logfile";
+}
+
 if ($ARGV[1] eq '-md5') {
 	$domd5=1;
 	print "MD5 Detected. Will do MD5 Digests as well...\n";
@@ -56,7 +62,7 @@ if ($ARGV[0] eq '-init'){
 	print "Check Detected: Now Checking File Sanity...\n";
 	&Go;
 } else {
-	print "\n\nViperDB v0.9.5\n";
+	print "\n\nViperDB v0.9.6\n";
 	print "ERROR: Unrecognized option or none given.\n";
 	print "usage: ViperDB [-init|-check|-checkstrict] [-md5]\n";
 	print "     -init	  Initializes the ViperDB Databases\n";
@@ -124,10 +130,12 @@ sub ParseConfigs
 			} elsif ($sub eq "rignore") {
 				push @rignore, $_;
 			} elsif ($sub eq "dirs") {
-				$cdirs{$_} = $_;
-				if (/^recursive\s*(.+)/) {
-					$rdirs{$1} = $1;
+				$tmp = $_;
+				if ($tmp =~ /^recursive\s*(.+)/) {
+					$tmp = $1;
+					$rdirs{$tmp} = $tmp;
 				}
+				$cdirs{$tmp} = $tmp;
 			} elsif ( $sub eq "options") {
 				# add some sanity check pls.
 				if (/^hostname\s*=\s*(.+)/) {
@@ -151,20 +159,30 @@ sub ParseConfigs
 					$timelock = $1;
 				} elsif (/^dbfname\s*=\s*(.+)/) {
 					$dbfname = $1;
+				} elsif (/^logfile\s*=\s*(.+)/) {
+					$logfile = $1;
 				}
 			}
 		}
 	}
 
 	$found = 1;
-	my %ndirs = {};
+	%ndirs = {};
 	my $file;
 	while ($found) {
 		$found = 0;
 		foreach $dir (sort values %rdirs) {
-			opendir (DIRLIST, $dir) || &Err ("internal", "Can't opendir $dir$!");
+			if (!defined $dir) {
+				# undef->undef gets passed
+				# anyone know how to efficiently COMPLETELY delete contents of a hash
+				next;
+			}
+			opendir (DIRLIST, $dir) || &Err ("internal", "Can't opendir $dir\n$!\nSkipping.");
 				my @files = readdir(DIRLIST);
 				foreach $file (@files) {
+					if ($file eq "." || $file eq "..") {
+						next;
+					}
 					# not this means when a subdir of a to be recursed directory is also
 					# in the config file, its recursion will be skipped. This is a "bug"
 					# in this algorithm, but can be used as a feature to exclude recusing
@@ -178,7 +196,7 @@ sub ParseConfigs
 					$fname =~ s/\/+/\//g;
 					if (-d $fname && !exists $cdirs{$fname} &&
 							! exists $ignore{"$file"}) {
-						$cdirs {$file} = $ndirs{$file} = $file;
+						$cdirs {$fname} = $ndirs{$fname} = $fname;
 						$found = 1;
 					}
 				}
@@ -218,7 +236,8 @@ sub Err
 sub OpenDB 
 {
 	my $wd = shift;
-	my $fname = "$wd/$dbfname";
+	#my $fname = "$wd/$dbfname";
+	my $fname = "$dbfname";
 	$dbopened == 1 && &CloseDB($wd);
 
 	if ($strict != 1 && $os_type ne 'sunos') {
@@ -234,28 +253,40 @@ sub OpenDB
 		if (-e $fname) {
 			$str = "";
 			if ($strict == 0) {
-				$str .= " Deleting.";
+				$str .= "Deleting.";
 				chmod(0600, $fname);
 				unlink $fname;
 			}
-			&Err ("critical", "In $wd, $fname isn't a file.\n$str");
+			&Err ("critical", "In $wd, $dbfname isn't a file.\n$str");
 		} elsif (!$init) {
 			$str = "";
 			if ($strict == 0) {
-				$str .= " Creating.";
+				$str .= "Creating.";
 			}
-			&Err ("suspicious", "In $wd, $fname doesn't exist.\n$str");
+			&Err ("suspicious", "In $wd, $dbfname doesn't exist.\n$str");
 		}
 	}
 	# create or open.
+	$dbopened = 1;
 	if ($strict) {
-		tie %db, 'GDBM_File', $fname, &GDBM_READER, 0400;
+		if (! tie (%db, 'GDBM_File', $fname, &GDBM_READER, 0400)) {
+			$dbopened = 0;
+			&Err ("critical", 
+				"In $wd, can't tie $dbfname\nDatabase probably corrupt, suggesting deleting.");
+		}
 	} else {
 		# if exists, chmod so we can write
 		( -f $fname ) && chmod (0600, $fname);
-		tie %db, 'GDBM_File', $fname, &GDBM_WRCREAT, 0600;
+		if (! tie (%db, 'GDBM_File', $fname, &GDBM_WRCREAT, 0600)) {
+			&Err ("strange", 
+				"In $wd, can't tie $dbfname\nDeleting and retrying...");
+			unlink $fname;
+			if (! tie (%db, 'GDBM_File', $fname, &GDBM_WRCREAT, 0600)) {
+				$dbopened = 0;
+				&Err ("critical", "Failed. Unable to monitor $wd.");
+			}
+		}
 	}
-	$dbopened = 1;
 }
 
 sub CloseDB 
@@ -267,7 +298,7 @@ sub CloseDB
 	untie %db;
 
 	if (! -f $fname) {
-		&Err ("critical", "In $wd, $fname isn't a file while closing.");
+		&Err ("critical", "In $wd, $dbfname isn't a file while closing.");
 	} else {
 		if ($strict == 0) {
 			chmod(0400, $fname);
@@ -334,7 +365,7 @@ sub ChkFile
 	my $file = shift;
 	my $fname = "$wd/$file";
 	$fname =~ s/\/+/\//g;
-	if ( -d $fname ) {
+	if ( -d $file ) {
 		return;
 	}
 	foreach $ig (@rignore) {
@@ -345,7 +376,7 @@ sub ChkFile
 			return;
 		}
 	}
-	if ( ! -e $fname && ! -l $fname) {
+	if ( ! -e $file && ! -l $file) {
 		&Err ("file deleted", "$fname");
 		$trouble++;
 		if (!$strict) {
@@ -353,19 +384,19 @@ sub ChkFile
 		}
 		return;
 	}
-	if ( -l $fname ) {
-		($dev, $ino, $mode, $nlink, $uid, $gid, $rdev, $size, $atime, $mtime, $ctime, $blksize, $blocks) = lstat $fname;
-		$size = readlink $fname;
+	if ( -l $file ) {
+		($dev, $ino, $mode, $nlink, $uid, $gid, $rdev, $size, $atime, $mtime, $ctime, $blksize, $blocks) = lstat $file;
+		$size = readlink $file;
 	} else {
-		($dev, $ino, $mode, $nlink, $uid, $gid, $rdev, $size, $atime, $mtime, $ctime, $blksize, $blocks) = stat $fname;
+		($dev, $ino, $mode, $nlink, $uid, $gid, $rdev, $size, $atime, $mtime, $ctime, $blksize, $blocks) = stat $file;
 	}
 	if (exists $db{"$file"}) {
 		@olddat = split /,/,$db{"$file"};
 	} else {
 		@olddat = (0,0,0,0,0,0);
 	}
-	$attr = &GetAttr ($fname);
-	$digest = &GetMD5 ($fname);
+	$attr = &GetAttr ($file);
+	$digest = &GetMD5 ($file);
 	if ($digest eq "file" && &FType ($olddat[6]) eq "file") {
 		$digest = $olddat[6];
 	}
@@ -374,6 +405,14 @@ sub ChkFile
 		&Err ("file added", "$fname");
 		$trouble++;
 		if (!$strict) {
+			# enforce MD5 check so we avoid double complaints.
+			$nft = &FType ($digest);
+			if (! $domd5 && $nft eq "file") {
+				$domd5 = 1;
+				$digest = &GetMD5 ($file);
+				$domd5 = 0;
+				$newdat = "$size,$mode,$attr,$uid,$gid,$mtime,$digest";
+			}
 			$db{$file} = $newdat;
 		}
 	} elsif ($newdat ne $db{"$file"}) {
@@ -388,6 +427,15 @@ sub ChkFile
 		# 6 - digest
 		$oft = &FType ($olddat[6]);
 		$nft = &FType ($digest);
+
+		# check MD5 now so we have more precise report and in case you combine MD5
+		# and non-MD5 check it doesn't complain twice
+		if (! $domd5 && $nft eq "file") {
+			$domd5 = 1;
+			$digest = &GetMD5 ($file);
+			$domd5 = 0;
+			$newdat = "$size,$mode,$attr,$uid,$gid,$mtime,$digest";
+		}
 		if ($oft ne $nft) {
 			&Err ("file type", "was $oft\nnow $nft");
 		} else {
@@ -395,8 +443,8 @@ sub ChkFile
 				if ($nft eq "symlink") {
 					$str = "";
 					if ($strict) {
-						unlink $fname;
-						symlink $fname, $olddat[0];
+						unlink $file;
+						symlink $file, $olddat[0];
 						$str = ", changed back";
 					}
 					&Err ("symlink", "was $olddat[0]\nnow $size$str");
@@ -407,10 +455,12 @@ sub ChkFile
 			if ($mode ne $olddat[1]) {
 				$str = "";
 				if ($strict) {
-					chmod $olddat[1], $fname;
+					chmod $olddat[1], $file;
 					$str = ", changed back";
 				}
-				&Err ("perms", "were $olddat[1]\n now $mode$str");
+				$omodestr = sprintf "%.4o", $olddat[1];
+				$nmodestr = sprintf "%.4o", $mode;
+				&Err ("perms", "were $omodestr\n now $nmodestr$str");
 			}
 			if ($attr ne $olddat[2]) {
 				$str = "";
@@ -423,7 +473,7 @@ sub ChkFile
 			if ($uid ne $olddat[3] || $gid ne $olddat[4]) {
 				$str = "";
 				if ($strict) {
-					chown $olddat[3], $olddat[4], $fname;
+					chown $olddat[3], $olddat[4], $file;
 					$str = ", changed back";
 				}
 				&Err ("uid/gid", "were $olddat[3]:$olddat[4]\n now $uid:$gid$str");
@@ -448,8 +498,12 @@ sub Go
 {
 	my $file;
 	foreach $wd (values %cdirs) {
+		chdir $wd;
 		&OpenDB($wd);
-		opendir(DIRLIST, $wd) || &Err ("internal", "Can't opendir $wd:\n$!");
+		if (!opendir(DIRLIST, $wd)) {
+			&Err ("internal", "Can't opendir $wd:\n$!\nSkipping.");
+			next;
+		}
 		my @files = readdir(DIRLIST);
 		FILE: foreach $file (@files) {
 			(exists $ignore{$file}) && next;
@@ -457,14 +511,15 @@ sub Go
 		}
 		close (DIRLIST);
 		foreach $file (keys %db) {
-			if (! -e "$wd/$file") {
+			if (! -e "$file") {
 				&ChkFile ($wd, $file);
 			}
 		}
 		&CloseDB($wd);
 	}
 	if ($trouble) {
-		&Err ("info", "END RUN - $trouble changed files detected.");
+		$str = &strftime ("%c", localtime);
+		&Err ("info", "$str\nEND RUN - $trouble changed files detected.");
 	}
 	close (LOG);
 	if (defined $email && $trouble && ! $init) {
